#include "Autopilot.h"




void pitch_controller::begin(double* _input,
	                         double* _output,
	                         double* _setpoint,
	                         const double& _p,
	                         const double& _i,
	                         const double& _d,
	                         const pOn& _pOn,
	                         const dir& _direction,
	                         const unsigned int& _minSamplePeriodMs,
	                         const double& _bias,
							 const double& _rollGain)
{
	input    = _input;
	output   = _output;
	setpoint = _setpoint;
	rollGain = _rollGain;
	setCoefficients(_p, _i, _d);
	setPOn(_pOn);
	setBias(_bias);
	setDirection(_direction);
	setSampleTime(_minSamplePeriodMs);

	reset();
	start();
}




void pitch_controller::compute(const double& roll)
{
	if (timer.fire() && modeType == ON)
	{
		kp = pIn;
		ki = iIn * (timer.timeDiff * 1000.0);
		kd = dIn / (timer.timeDiff * 1000.0);

		if (direction == BACKWARD)
		{
			kp *= -1;
			ki *= -1;
			kd *= -1;
		}

		lastInput = curInput;
		lastSetpoint = curSetpoint;
		lastError = curError;

		curInput = *input;
		curSetpoint = *setpoint;
		curError = curSetpoint - curInput;

		double dInput = *input - lastInput;

		if (pOnType == P_ON_E)
			pOut = kp * curError;
		else if (pOnType == P_ON_M)
			pOut = -kp * dInput;

		iOut += (ki * ((curError + lastError) / 2.0)); // Trapezoidal integration
		iOut = constrain(iOut, windupMin, windupMax);  // Prevent integral windup

		dOut = -kd * dInput; // Derrivative on measurement

		double newOutput = bias + ((pOut + iOut + dOut) * roll_compensation(roll));
		newOutput = constrain(newOutput, outputMin, outputMax);
		*output = newOutput;
	}
}




double pitch_controller::roll_compensation(const double& roll)
{
	return 1 + (rollGain * abs(tan(radians(roll))));
}




void pitch_controller::setRollGain(const double& _rollGain)
{
	rollGain = _rollGain;
}
